import typing
from . import BaseWidget
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QTableWidget, QAbstractItemView, QTableWidgetItem
from ...model import WidgetStringItem, WidgetIndexItem
from ...model.types import StringListItem
from ..interfaces import ExcitedSignalInterface


class TableWidget(QTableWidget,
                  BaseWidget,
                  ExcitedSignalInterface,
                  WidgetStringItem,
                  WidgetIndexItem,
                  StringListItem
                  ):
    _header_labels: list

    def set_index_value(self, value: int):
        self.selectRow(value or 0)

    def get_index_value(self) -> int:
        return self.currentRow()

    def get_string_list_value(self) -> typing.List[str]:
        if self.index.value < 0:
            return []
        return [self.item(self.index.value, col).text() for col in range(self.column_count)]

    def set_string_list_value(self, value: typing.List[typing.Any]):
        if not isinstance(value, list):
            raise ValueError('Only Support List Type Value')
        if len(value) != self.column_count:
            raise ValueError('Value Length Must Equal to TableWidget Column Count')

        self.row_count += 1
        for col in range(self.column_count):
            table_item = QTableWidgetItem(str(value[col]))
            table_item.setTextAlignment(Qt.AlignCenter)
            self.setItem(self.row_count - 1, col, table_item)
        self.index.value = self.row_count - 1
        self.auto_resize_column_width()
        self.string_list.check_change()

    @property
    def row_count(self):
        return self.rowCount()

    @row_count.setter
    def row_count(self, value: int):
        self.setRowCount(value)

    @property
    def column_count(self):
        return self.columnCount()

    @column_count.setter
    def column_count(self, value: int):
        self.setColumnCount(value)

    def set_changed_connection(self):
        # noinspection PyUnresolvedReferences
        self.currentCellChanged.connect(self.index.check_change)

    def set_excited_signal_connection(self):
        # noinspection PyUnresolvedReferences
        self.doubleClicked.connect(lambda *_: self.excited.emit())

    def set_just_show_mode(self):
        self.auto_resize = True
        self.verticalHeader().hide()
        self.resizeColumnsToContents()
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)

    def set_select_rows_mode(self):
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setStyleSheet("selection-background-color: lightBlue;selection-color: black;")
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)

    @property
    def selected_rows(self) -> list:
        return sorted(list({self.indexFromItem(item).row() for item in self.selectedItems()}))

    @property
    def auto_resize(self):
        return getattr(self, 'resize', False)

    @auto_resize.setter
    def auto_resize(self, value: bool):
        setattr(self, 'resize', value)

    @property
    def header_labels(self) -> typing.List[str]:
        return self._header_labels or []

    @header_labels.setter
    def header_labels(self, labels: list):
        self.setColumnCount(len(labels))
        self.setHorizontalHeaderLabels(labels)
        self._header_labels = labels

    def auto_resize_column_width(self):
        if self.auto_resize:
            self.resizeColumnsToContents()
            col_count = self.columnCount()
            col_width = sum(list([self.columnWidth(i) for i in range(col_count)]))
            if col_width < self.width():
                for i in range(col_count):
                    self.setColumnWidth(i, self.columnWidth(i) / col_width * self.width())

    def resizeEvent(self, event):
        super(TableWidget, self).resizeEvent(event)
        self.auto_resize_column_width()



import typing
from . import BaseWidget
from ...gui import Painter
from ...core import EventLoop
from PySide2.QtGui import QCloseEvent, QColor
from ..interfaces import ClosedSignalInterface, ClassExecInterface


class Widget(BaseWidget, ClosedSignalInterface, ClassExecInterface):
    def __init__(self, parent: typing.Optional['Widget'] = None, *args):
        super().__init__(parent, *args)

    def closeEvent(self, event: QCloseEvent):
        self.deal_with_close(event)

    def exec(self):
        with EventLoop() as event:
            self.show()
            self.closed.connect(event.quit)

    @property
    def background_color(self) -> typing.Optional[QColor]:
        return self.attach(lambda: None)

    @background_color.setter
    def background_color(self, value: QColor):
        self.assign(value)
        self.update()

    def paintEvent(self, *args, **kwargs):
        painter = Painter(self)

        if self.background_color is not None:
            painter.fillRect(self.rect(), self.background_color)

        self.paint(painter)
        painter.end()

    def paint(self, painter: Painter):
        pass



import typing
from . import BaseWidget
from PySide2.QtWidgets import QListWidget
from ...model import WidgetStringItem, WidgetIndexItem
from ...model.types import StringListItem
from ..interfaces import ExcitedSignalInterface


class ListWidget(QListWidget,
                 BaseWidget,
                 ExcitedSignalInterface,
                 WidgetStringItem,
                 WidgetIndexItem,
                 StringListItem
                 ):
    def set_excited_signal_connection(self):
        # noinspection PyUnresolvedReferences
        self.doubleClicked.connect(lambda *_: self.excited.emit())

    def get_string_value(self) -> typing.Optional[str]:
        if self.index.value >= 0:
            return self.currentItem().text()
        else:
            return None

    def set_string_value(self, value: str):
        texts = self.string_list.value
        assert isinstance(texts, list)

        if value in texts:
            self.index.value = texts.index(value)
        else:
            raise ValueError(f'Value {value} Not Found in List')

    def set_changed_connection(self):
        # noinspection PyUnresolvedReferences
        self.currentTextChanged.connect(self.string.check_change)
        # noinspection PyUnresolvedReferences
        self.currentRowChanged.connect(self.index.check_change)

    def get_index_value(self) -> int:
        return self.currentRow()

    def set_index_value(self, value: int):
        self.setCurrentRow(value or 0)

    def get_string_list_value(self) -> typing.List[str]:
        return [self.item(index).text() for index in range(self.count())]

    def set_string_list_value(self, value: typing.List[str]):
        value = value or []
        self.clear()
        self.addItems(value)
        self.string_list.check_change()



from . import BaseWidget
from ...model import WidgetStringItem
from PySide2.QtWidgets import QPushButton
from ..interfaces import ExcitedSignalInterface


class PushButton(QPushButton, BaseWidget, ExcitedSignalInterface, WidgetStringItem):
    def set_excited_signal_connection(self):
        # noinspection PyUnresolvedReferences
        self.clicked.connect(self.excited.emit)

    def get_value(self) -> str:
        return self.text()

    def set_value(self, value: str):
        self.setText(value)
        self.check_change()



import typing
from pathlib import Path
from ...model import base
from ..layouts import SquareLayout
from PySide2.QtCore import QPoint
from PySide2.QtGui import QPixmap
from PySide2.QtGui import QPicture
from PySide2.QtGui import QPainter
from PySide2.QtWidgets import QWidget
from PySide2.QtSvg import QSvgGenerator
from PySide2.QtWidgets import QDockWidget
from PySide2.QtWidgets import QMainWindow
from PySide2.QtWidgets import QHBoxLayout
from PySide2.QtPrintSupport import QPrinter


class BaseWidget(QWidget, base.AttachMixin):
    @property
    def center_widget(self) -> typing.Optional['BaseWidget']:
        return self.attach(lambda: None)

    @center_widget.setter
    def center_widget(self, widget: 'BaseWidget'):
        self.assign(widget)

    def set_central_widget(self, widget: 'BaseWidget', spacing: int = 0, layout_type=QHBoxLayout):
        if self.center_widget is not None:
            self.layout().replaceWidget(self.center_widget, widget)
            self.center_widget.deleteLater()
            self.center_widget = widget
            return

        if isinstance(self, QMainWindow):
            self.setCentralWidget(widget)
        elif isinstance(self, QDockWidget):
            self.setWidget(widget)
        else:
            layout = layout_type()
            layout.setSpacing(spacing)
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(widget)
            self.setLayout(layout)

        self.center_widget = widget

    def set_square_widget(self, widget: QWidget, spacing: int = 0):
        self.set_central_widget(widget, spacing, layout_type=SquareLayout)

    def set_focus(self):
        return self.setFocus()

    def set_enabled(self):
        return self.setEnabled(True)

    def set_disabled(self):
        return self.setEnabled(False)

    @property
    def enabled(self) -> bool:
        return self.isEnabled()

    @enabled.setter
    def enabled(self, value: bool):
        if value:
            self.set_enabled()
        else:
            self.set_disabled()

    def export_to_pdf(self: 'BaseWidget', filename: Path):
        width_a4 = 730
        height_a4 = 1060

        is_landscape = self.width() > self.height()
        if is_landscape:
            self.resize(height_a4, width_a4)  # landscape
        else:
            self.resize(width_a4, height_a4)  # portrait

        # render widget to picture
        picture = QPicture()
        painter = QPainter(picture)
        self.render(painter, QPoint(0, 0))
        painter.end()

        # set up PDF printer
        printer = QPrinter()
        printer.setOutputFormat(QPrinter.PdfFormat)
        printer.setOutputFileName(str(filename))
        if is_landscape:
            printer.setOrientation(QPrinter.Landscape)

        # draw picture on printer
        painter = QPainter()
        ok = painter.begin(printer)
        if ok:
            painter.drawPicture(0, 0, picture)
            ok = painter.end()
        return ok

    def export_to_picture(self: 'BaseWidget', filename: Path):
        picture = QPixmap(self.width(), self.height())
        painter = QPainter(picture)
        self.render(painter, QPoint(0, 0))
        painter.end()

        return picture.save(str(filename))

    def export_to_svg(self, filename: Path) -> bool:
        generator = QSvgGenerator()
        generator.setFileName(str(filename))
        generator.setSize(self.size())
        generator.setViewBox(self.rect())
        self.render(generator)
        return True



from . import layouts
from . import interfaces
from . import widgets

from .widgets import Widget
from .widgets import Dialog
from .widgets import MainWindow
from .widgets import GroupBox
from .widgets import DockWidget

from .widgets import Label
from .widgets import LineEdit

from .widgets import SpinBox
from .widgets import DoubleSpinBox

from .widgets import ComboBox
from .widgets import ListWidget
from .widgets import TableWidget

from .widgets import InputDialog

from .widgets import Action
from .widgets import Shortcut
from .widgets import PushButton



from ... import SignalSender
from ...model.base import AttachMixin


class ExcitedSignalInterface(AttachMixin):
    @property
    def excited(self) -> SignalSender:
        return self.attach(creator=SignalSender, after_created=self.set_excited_signal_connection)

    def set_excited_signal_connection(self):
        raise NotImplementedError



from ... import SignalSender
from PySide2.QtGui import QCloseEvent
from ...model.base import AttachMixin


class ClosedSignalInterface(AttachMixin):
    @property
    def closed(self) -> SignalSender:
        return self.attach(SignalSender)

    @property
    def cannot_closed(self) -> SignalSender:
        return self.attach(SignalSender)

    @property
    def can_close(self) -> bool:
        return self.attach(lambda: True)

    @can_close.setter
    def can_close(self, value: bool):
        self.assign(value)

    def deal_with_close(self, event: QCloseEvent):
        if self.can_close:
            self.closed.emit()
            event.accept()
        else:
            self.cannot_closed.emit()
            event.ignore()



from .class_exec_interface import ClassExecInterface
from .closed_signal_interface import ClosedSignalInterface
from .excited_signal_interface import ExcitedSignalInterface
from .message_box_interface import MessageBoxInterface



from PySide2.QtWidgets import QMessageBox


class MessageBoxInterface:
    def question(self, text, title='选择'):
        # noinspection PyCallByClass
        return QMessageBox.question(self, title, text, QMessageBox.Yes | QMessageBox.No) == QMessageBox.Yes

    def warning(self, text, title='警告'):
        # noinspection PyCallByClass
        QMessageBox.warning(self, title, text)

    def information(self, text, title='提示'):
        # noinspection PyCallByClass
        QMessageBox.information(self, title, text)

    def about(self, text, title='关于'):
        # noinspection PyCallByClass
        QMessageBox.about(self, title, text)

    def message(self, ok=True, ok_msg='成功', bad_msg='失败'):
        if ok:
            self.information(ok_msg)
        else:
            self.warning(bad_msg)



import math
import typing
import operator
from PySide2.QtCore import QSizeF


class SizeF(QSizeF):
    def __init__(self, w: float = 0, h: float = 0):
        super().__init__(w, h)

    @property
    def w(self) -> float:
        return self.width()

    @property
    def h(self) -> float:
        return self.height()

    @property
    def horizontal(self) -> 'SizeF':
        return SizeF(self.w, 0)

    @property
    def vertical(self) -> 'SizeF':
        return SizeF(0, self.h)

    def __calc__(self, other: typing.Union[int, float, 'SizeF'], op: typing.Callable[[float, float], float]):
        if isinstance(other, SizeF):
            return SizeF(
                op(self.w, other.w),
                op(self.h, other.h)
            )
        else:
            return SizeF(
                op(self.w, other),
                op(self.h, other)
            )

    def __add__(self, other: typing.Union[int, float, 'SizeF']):
        return self.__calc__(other, operator.add)

    __radd__ = __add__

    def __sub__(self, other: typing.Union[int, float, 'SizeF']):
        return self.__calc__(other, operator.sub)

    __rsub__ = None

    def __mul__(self, other: typing.Union[int, float, 'SizeF']):
        return self.__calc__(other, operator.mul)

    __rmul__ = __mul__

    def __truediv__(self, other: typing.Union[int, float, 'SizeF']):
        return self.__calc__(other, operator.truediv)

    def __neg__(self):
        return SizeF(-self.w, -self.h)

    def __eq__(self, other: 'SizeF'):
        return math.isclose(self.w, other.w) and math.isclose(self.h, other.h)



class ClassExecInterface:
    def exec(self):
        raise NotImplementedError

    @classmethod
    def class_exec(cls, *args, **kwargs):
        # noinspection PyArgumentList
        cls(*args, **kwargs).exec()  # pragma: no cover



from . import SignalSender
from PySide2.QtCore import QTimer


class Timer(QTimer):
    def __init__(self, single_shot: bool = False):
        super().__init__()
        self.setSingleShot(single_shot)

        # replace original Signal to SignalSender, trick of type
        self.triggered = SignalSender()
        self.timeout.connect(self.triggered.emit)
        self.timeout = self.triggered

    def start(self, seconds: float):
        super().start(int(seconds * 1000))



from ..core import SizeF
from PySide2.QtGui import QFont, QFontMetrics


def text_size(text, font: QFont) -> 'SizeF':
    font_metrics = QFontMetrics(font)

    lines = text.split('\n')
    width = max(map(font_metrics.width, lines))
    height = font_metrics.height() * len(lines)
    return SizeF(width, height)


def __get_scaling_ratio__():
    base = 800.0  # the width of a character when dpi = 96
    width = text_size('Q', QFont('Courier New', 1000)).w
    return width / base


scaling_ratio = __get_scaling_ratio__()



import typing
from . import BaseWidget
from PySide2.QtWidgets import QComboBox
from ...model import WidgetStringItem, WidgetIndexItem
from ...model.types import StringListItem


class ComboBox(QComboBox,
               BaseWidget,
               WidgetStringItem,
               WidgetIndexItem,
               StringListItem
               ):
    def get_string_value(self) -> str:
        return self.currentText()

    def set_string_value(self, value: typing.Optional[str]):
        texts = self.string_list.value
        assert isinstance(texts, list)

        if value in texts:
            self.index.value = texts.index(value)
        else:
            raise ValueError(f'Value {value} Not Found in List')

    def set_changed_connection(self):
        # noinspection PyUnresolvedReferences
        self.currentIndexChanged[int].connect(self.index.check_change)
        # noinspection PyUnresolvedReferences
        self.currentIndexChanged[str].connect(self.string.check_change)

    def get_index_value(self) -> int:
        return self.currentIndex()

    def set_index_value(self, value: int):
        if isinstance(value, int):
            value = max(0, min(value, self.count() - 1))
            self.setCurrentIndex(value)

    def get_string_list_value(self) -> typing.List[str]:
        return [self.itemText(index) for index in range(self.count())]

    def set_string_list_value(self, value: typing.List[str]):
        value = value or []

        old_count = self.count()
        new_count = len(value)
        for i in range(min(old_count, new_count)):
            self.setItemText(i, value[i])
        for index in reversed(range(new_count, old_count)):
            self.removeItem(index)
        self.addItems(value[old_count:new_count])

        self.index.check_change()
        self.string.check_change()
        self.string_list.check_change()



from . import BaseWidget
from PySide2.QtWidgets import QDialog
from PySide2.QtGui import QCloseEvent
from ..interfaces import ClosedSignalInterface, ClassExecInterface, MessageBoxInterface


class Dialog(QDialog, BaseWidget, ClosedSignalInterface, ClassExecInterface, MessageBoxInterface):
    def closeEvent(self, event: QCloseEvent):
        self.deal_with_close(event)

    def exec(self):
        super().exec_()



from ..base import ValueInterface
from ..base import ChangedValueEntity
from ..base import AbstractItem, AbstractProperty


class IntegerTypeMixin(ValueInterface):
    @property
    def value(self) -> int:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class IntegerTypeProperty(AbstractProperty, IntegerTypeMixin):
    pass


class IntegerPropertyInterface(ChangedValueEntity):
    @property
    def integer(self) -> IntegerTypeProperty:
        return self.attach(creator=IntegerTypeProperty, args=(
            self.get_integer_value, self.set_integer_value, self.changed
        ))

    def get_integer_value(self) -> int:
        return self.get_value()

    def set_integer_value(self, value: int):
        return self.set_value(value)


class IntegerItem(AbstractItem, IntegerPropertyInterface):
    pass
